var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { NativeModules, DeviceEventEmitter } from "react-native";
import { ledgerUSBVendorId, identifyUSBProductId } from "@ledgerhq/devices";
import { DisconnectedDeviceDuringOperation, DisconnectedDevice } from "@ledgerhq/errors";
import { log } from "@ledgerhq/logs";
import Transport from "@ledgerhq/hw-transport";
import { Subject, from, concat } from "rxjs";
import { mergeMap } from "rxjs/operators";
const disconnectedErrors = [
    "I/O error",
    "Attempt to invoke virtual method 'int android.hardware.usb.UsbDevice.getDeviceClass()' on a null object reference",
    "Invalid channel",
    "Permission denied by user for device",
];
const listLedgerDevices = () => __awaiter(void 0, void 0, void 0, function* () {
    const devices = yield NativeModules.HID.getDeviceList();
    return devices.filter(d => d.vendorId === ledgerUSBVendorId);
});
const liveDeviceEventsSubject = new Subject();
DeviceEventEmitter.addListener("onDeviceConnect", (device) => {
    if (device.vendorId !== ledgerUSBVendorId)
        return;
    const deviceModel = identifyUSBProductId(device.productId);
    liveDeviceEventsSubject.next({
        type: "add",
        descriptor: device,
        deviceModel,
    });
});
DeviceEventEmitter.addListener("onDeviceDisconnect", (device) => {
    if (device.vendorId !== ledgerUSBVendorId)
        return;
    const deviceModel = identifyUSBProductId(device.productId);
    liveDeviceEventsSubject.next({
        type: "remove",
        descriptor: device,
        deviceModel,
    });
});
const liveDeviceEvents = liveDeviceEventsSubject;
/**
 * Ledger's React Native HID Transport implementation
 * @example
 * import TransportHID from "@ledgerhq/react-native-hid";
 * ...
 * TransportHID.create().then(transport => ...)
 */
class HIDTransport extends Transport {
    constructor(nativeId, productId) {
        super();
        this.id = nativeId;
        this.deviceModel = identifyUSBProductId(productId);
    }
    /**
     * List currently connected devices.
     * @returns Promise of devices
     */
    static list() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!NativeModules.HID)
                return Promise.resolve([]);
            return yield listLedgerDevices();
        });
    }
    /**
     * Listen to ledger devices events
     */
    static listen(observer) {
        if (!NativeModules.HID)
            return {
                unsubscribe: () => { },
            };
        return concat(from(listLedgerDevices()).pipe(mergeMap(devices => from(devices.map(device => ({
            type: "add",
            descriptor: device,
            deviceModel: identifyUSBProductId(device.productId),
        }))))), liveDeviceEvents).subscribe(observer);
    }
    /**
     * Open a the transport with a Ledger device
     */
    static open(deviceObj) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const nativeObj = yield NativeModules.HID.openDevice(deviceObj);
                return new HIDTransport(nativeObj.id, deviceObj.productId);
            }
            catch (error) {
                if (disconnectedErrors.includes(error.message)) {
                    throw new DisconnectedDevice(error.message);
                }
                throw error;
            }
        });
    }
    /**
     * @param {*} apdu input value
     * @returns Promise of apdu response
     */
    exchange(apdu) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.exchangeAtomicImpl(() => __awaiter(this, void 0, void 0, function* () {
                try {
                    const apduHex = apdu.toString("hex");
                    log("apdu", "=> " + apduHex);
                    const resultHex = yield NativeModules.HID.exchange(this.id, apduHex);
                    const res = Buffer.from(resultHex, "hex");
                    log("apdu", "<= " + resultHex);
                    return res;
                }
                catch (error) {
                    if (disconnectedErrors.includes(error.message)) {
                        this.emit("disconnect", error);
                        throw new DisconnectedDeviceDuringOperation(error.message);
                    }
                    throw error;
                }
            }));
        });
    }
    /**
     * Close the transport
     * @returns Promise
     */
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.exchangeBusyPromise;
            return NativeModules.HID.closeDevice(this.id);
        });
    }
    setScrambleKey() { }
}
/**
 * Check if the transport is supported (basically true on Android)
 */
HIDTransport.isSupported = () => Promise.resolve(!!NativeModules.HID);
export default HIDTransport;
//# sourceMappingURL=index.js.map